use crate::{gate::Gate, gate_type::GateType};

#[derive(Debug, Clone)]
pub struct Circuit {
    pub input_bits: Vec<bool>, // Input wires
    pub gates: Vec<Gate>,      // All gates
}

impl Circuit {
    pub fn eval(self) -> bool {
        let mut evaluated_gates = vec![];

        for gate in self.gates {
            let result = gate.eval(&self.input_bits, &evaluated_gates);
            evaluated_gates.push(result);
        }

        match evaluated_gates.pop() {
            Some(result) => result,
            None => panic!("Bruh moment"),
        }
    }

    /*
        This method should create a circuit that outputs 1 if the first number A (encoded in the first n
        bits) is greater than the second number B (encoded in the next n bits) .
    */
    pub fn compare_n_bit_numbers(input_bits: Vec<bool>, n: usize) -> Self {
        if input_bits.len() < 2 * n {
            panic!(
                "Expected input_bits to be of at least length {}, but it was {}",
                2 * n,
                input_bits.len()
            )
        }

        /*
            base case n=1: 1-bit:

                A B | A > B
                -----------
                0 0 |   0
                0 1 |   0
                1 0 |   1
                1 1 |   0

            n+1 case: 2-bit:

                A1 A0  B1 B0  | A > B | (A1 > B1) || ((A1 == B1) && (A0 > B0))
                ----------------------|------------------------------------
                0  0   0  0   |   0   |            0
                0  0   0  1   |   0   |            0
                0  0   1  0   |   0   |            0
                0  0   1  1   |   0   |            0
                0  1   0  0   |   1   |            1
                0  1   0  1   |   0   |            0
                0  1   1  0   |   0   |            0
                0  1   1  1   |   1   |            1
                1  0   0  0   |   1   |            1
                1  0   0  1   |   1   |            1
                1  0   1  0   |   0   |            0
                1  0   1  1   |   0   |            0
                1  1   0  0   |   1   |            1
                1  1   0  1   |   1   |            1
                1  1   1  0   |   1   |            1
                1  1   1  1   |   0   |            0

            The inductive pattern should become apparent now.
            For illustration here's the case for n=4, which should show the
            recursive characteristics of the formula.

            (A3 > B3) ||
            ((A3 == B3) && (A2 > B2)) ||
            ((A3 == B3) && (A2 == B2) && (A1 > B1)) ||
            ((A3 == B3) && (A2 == B2) && (A1 == B1)) && (A0 > B0))

        */

        let gates = create_n_bit_comparator_gates(n);
        return Circuit { input_bits, gates };
    }
}

fn create_n_bit_comparator_gates(n: usize) -> Vec<Gate> {
    let mut all_gates: Vec<Gate> = Vec::with_capacity(3*n);
    let mut and_gate_indices: Vec<usize> = vec![0; n];
    let mut eq_gate_indices: Vec<usize> = Vec::with_capacity(n-1);

    const EMPTY_VEC: Vec<usize> = Vec::new();

    for curr in 0..n {
        // Gate(A_curr > B_curr)
        let mut and_gate_input_indices = vec!(all_gates.len());
        println!("GT = {}", all_gates.len());
        all_gates.push(Gate::new(GateType::Bigger, EMPTY_VEC, vec![curr, curr + n]));

        // Bit to the left of curr. The one at array-index 0 doesn't have one.
        if curr != 0 {
            // Gate(A_curr-1 = B_curr-1)
            eq_gate_indices.push(all_gates.len());
            println!("EQ = {}", all_gates.len());
            all_gates.push(Gate::new(GateType::Equal, EMPTY_VEC, vec![curr - 1, curr - 1 + n]));
        }

        and_gate_input_indices.extend(eq_gate_indices.iter());

        // The AND spanning all gates for this bit
        and_gate_indices.push(all_gates.len());
        println!("&& = {}", all_gates.len());
        all_gates.push(Gate::new(GateType::And, and_gate_input_indices, EMPTY_VEC));
    }

    // the OR spanning all ANDs
    let or_gate = Gate::new(GateType::Or, and_gate_indices, EMPTY_VEC);
    all_gates.push(or_gate);

    return all_gates;
}
